/* $Id: parse.c 27 2023-03-31 20:28:48Z nishi $ */
#include <nikkel/parse.h>

#include <nikkel/config.h>
#include <nikkel/util.h>
#include <nikkel.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <dirent.h>
#include <sys/utsname.h>
#include <unistd.h>

struct nikkel_entry {
	char* name;
	char* original_name;
	char* prefix;
	char* entryprefix;
	char letter;
};

struct nikkel_section {
	char* name;
	char** subsections;
};

#define CHECK_NULL	if(value == NULL){\
				free(result->p_title);\
				char* err = malloc(513);\
				memset(err, 0, 513);\
				sprintf(err, "<span class=\"special-message\"><span class=\"color-red\">&emsp;</span><span class=\"special-error\">ERROR</span>: Line number %d: `%s` needs the argument<span class=\"color-red\">&emsp;</span></span>", line, key);\
				result->p_title = sstrdup("Error");\
				char* tmp_body = result->p_body;\
				result->p_body = sstrcat(tmp_body, err);\
				free(tmp_body);\
				free(err);\
				break;\
			}

struct nikkel_parsed* nikkel_parse(struct nikkel_config* config, FILE* f, const char* content){
	struct nikkel_parsed* result = malloc(sizeof(*result));
	result->p_title = sstrdup("No Title");
	int counter = 0;
	char ch;
	char* buffer = malloc(1);
	buffer[0] = 0;
	result->p_body = malloc(1);
	result->p_pagetitle = NULL;
	result->p_body[0] = 0;
	result->p_category = malloc(sizeof(*result->p_category));
	result->p_category[0] = NULL;
	result->p_forbid = 0;
	char* table_width = NULL;
	int category_count = 0;
	int line = 0;
	char code = 0;
	char* image_width = NULL;
	int i = 0;
	int section_number = 0;
	int subsection_number = 1;
	struct nikkel_section** sections = malloc(sizeof(*sections));
	sections[0] = NULL;
	enum TYPES {
		TABLE_NONE = 0,
		TABLE_HEADER,
		TABLE_DATA
	};
	char* types = malloc(1);
	types[0] = TABLE_NONE;
	int type_count = 0;
	int table_place = TABLE_LEFT;
	char forbid = 0;
	result->p_section = SECTION_RIGHT;
	while(1){
		if(f == NULL){
			ch = content[i];
			i++;
		}else{
			fread(&ch, 1, 1, f);
		}
		if((f == NULL ? ch == 0 : feof(f)) || ch == '\n'){
			line++;
			if(counter != 0){
				int i;
				char sbracket = 0;
				char* ch = malloc(2);
				ch[0] = 0;
				ch[1] = 0;
				char* sbracket_buffer = malloc(1);
				sbracket_buffer[0] = 0;
				int start_sbracket = 0;
				int bracket_count = 0;
				for(i = 0; buffer[i] != 0; i++){
					ch[0] = buffer[i];
					if(sbracket == 1){
						if(ch[0] == ']' && bracket_count == 0){
							char assign = 0;
							int j;
							for(j = start_sbracket; j < i; j++){
								if(buffer[j] == '|'){
									assign = 1;
									break;
								}
							}
							{
								char* orig_buffer = sstrdup(buffer);
								char* key = buffer + start_sbracket + 1;
								key[j - 1 - start_sbracket] = 0;
								char* value = NULL;
								if(assign == 1){
									value = sstrdup(buffer + j + 1);
									value[i - start_sbracket - sstrlen(key) - 2] = 0;
								}
								if(code == 1){
									if(strcmp(key, "/Code") == 0){
										code = 0;
										char* tmp = result->p_body;
										result->p_body = sstrcat(tmp, "</code></pre>");
										free(tmp);
										if(value != NULL) free(value);
										goto clean_sbracket;
									}
								}
								if(strcmp(key, "Title") == 0){
									CHECK_NULL;
									free(result->p_title);
									result->p_title = sstrdup(value);
								}else if(strcmp(key, "PageTitle") == 0){
									CHECK_NULL;
									if(result->p_pagetitle != NULL) free(result->p_pagetitle);
									result->p_pagetitle = sstrdup(value);
								}else if(strcmp(key, "Section") == 0){
									CHECK_NULL;
									subsection_number = 0;
									struct nikkel_section* sect = malloc(sizeof(*sect));
									sect->name = sstrdup(value);
									sect->subsections = malloc(sizeof(*(sect->subsections)));
									sect->subsections[0] = NULL;
									sections = realloc(sections, sizeof(*sections) * (section_number + 2));
									sections[section_number] = sect;
									sections[section_number + 1] = NULL;
									char* tmp = result->p_body;
									char* replaced = sreplace(value, ' ', '-');
									char* replaced_num = malloc(513);
									memset(replaced_num, 0, 513);
									sprintf(replaced_num, "%d-%d-", section_number + 1, 0);
									char* start = sstrcat4("<h3 id=\"section-", replaced_num, replaced, "\">");
									char* section = malloc(513);
									memset(section, 0, 513);
									sprintf(section, "%d. ", section_number + 1);
									char* fmt = sstrcat(section, value);
									free(section);
									char* more = sstrcat5(fmt, " <a href=\"#section-", replaced_num, replaced, "\">&#167;</a>");
									section_number++;
									free(replaced_num);
									free(fmt);
									result->p_body = sstrcat4(tmp, start, more, "</h3>\n<hr>\n");
									free(more);
									free(start);
									free(replaced);
									free(tmp);
								}else if(strcmp(key, "Subsection") == 0){
									CHECK_NULL;
									if(section_number == 0){
										char* tmp = result->p_body;
										char* err = malloc(513);
										memset(err, 0, 513);
										sprintf(err, "<span class=\"special-message\"><span class=\"color-yellow\">&emsp;</span><span class=\"special-warn\">WARN</span>: Line number %d: `Subsection` must be used after the `Section`. Ignored.<span class=\"color-yellow\">&emsp;</span></span><br>", line);
										result->p_body = sstrcat(tmp, err);
										free(tmp);
										free(err);
										goto clean_sbracket;
									}
									sections[section_number - 1]->subsections = realloc(sections[section_number - 1]->subsections, sizeof(char*) * (subsection_number + 2));
									sections[section_number - 1]->subsections[subsection_number] = sstrdup(value);
									sections[section_number - 1]->subsections[subsection_number + 1] = NULL;
									char* tmp = result->p_body;
									char* replaced = sreplace(value, ' ', '-');
									char* replaced_num = malloc(513);
									memset(replaced_num, 0, 513);
									sprintf(replaced_num, "%d-%d-", section_number, subsection_number + 1);
									char* start = sstrcat4("<h3 id=\"section-", replaced_num, replaced, "\">");
									char* section = malloc(513);
									memset(section, 0, 513);
									sprintf(section, "%d.%d. ", section_number, subsection_number + 1);
									char* fmt = sstrcat(section, value);
									free(section);
									char* more = sstrcat5(fmt, " <a href=\"#section-", replaced_num, replaced, "\">&#167;</a>");
									free(replaced_num);
									subsection_number++;
									free(fmt);
									result->p_body = sstrcat4(tmp, start, more, "</h3>\n<hr>\n");
									free(more);
									free(start);
									free(replaced);
									free(tmp);
								}else if(strcmp(key, "ImageWidth") == 0){
									if(value == NULL){
										if(image_width != NULL) free(image_width);
										image_width = NULL;
									}else{
										if(image_width != NULL) free(image_width);
										image_width = sstrdup(value);
									}
								}else if(strcmp(key, "Link") == 0 || strcmp(key, "Entry") == 0 || strcmp(key, "Image") == 0 || strcmp(key, "CategoryEntry") == 0){
									CHECK_NULL;
									char* link = malloc(1);
									link[0] = 0;
									char* name = malloc(1);
									name[0] = 0;
									int index = -1;
									for(j = 0; value[j] != 0; j++){
										if(value[j] == '|'){
											index = j;
											break;
										}
									}
									if(index == -1){
										free(link);
										link = sstrdup(value);
									}else{
										free(link);
										free(name);
										value[index] = 0;
										link = sstrdup(value);
										name = sstrdup(value + index + 1);
									}
									if(strcmp(key, "Entry") == 0){
										char* tmp = link;
										link = sstrcat3(config->c_prefix, "/", link);
										free(tmp);
									}else if(strcmp(key, "CategoryEntry") == 0){
										char* tmp = link;
										link = sstrcat3(config->c_prefix, "/Category:", link);
										free(tmp);
									}
									if(strcmp(key, "Image") == 0){
										char* tmp = result->p_body;
										result->p_body = sstrcat(tmp, "<span class=\"image-container\"");
										free(tmp);
										if(image_width != NULL){
											tmp = result->p_body;
											result->p_body = sstrcat4(tmp, " style=\"width: ", image_width, ";");
											free(tmp);
											if(name[0] != 0){
												tmp = result->p_body;
												result->p_body = sstrcat(tmp, "padding-right: 20px;");
												free(tmp);
											}
											tmp = result->p_body;
											result->p_body = sstrcat(tmp, "\"");
											free(tmp);

										}
										tmp = result->p_body;
										result->p_body = sstrcat4(tmp, "><a href=\"", link, "\">");
										free(tmp);

										tmp = result->p_body;
										result->p_body = sstrcat4(tmp, "<img src=\"", link, "\"");
										free(tmp);
										if(name[0] != 0){
											tmp = result->p_body;
											result->p_body = sstrcat4(tmp, " alt=\"", name, "\"");
											free(tmp);
										}
										if(image_width != NULL){
											tmp = result->p_body;
											result->p_body = sstrcat(tmp, " width=\"100%\"");
											free(tmp);
										}
										tmp = result->p_body;
										result->p_body = sstrcat4(tmp, "></a><div>", name, "</div></span>");
										free(tmp);
									}else{
										char* tmp = result->p_body;
										result->p_body = sstrcat4(tmp, "<a href=\"", link, "\">");
										free(tmp);
										tmp = result->p_body;
										result->p_body = sstrcat3(tmp, (index == -1 ? link : name), "</a>");
										free(tmp);
									}
									free(name);
									free(link);
								}else if(strcmp(key, "Category") == 0){
									CHECK_NULL;
									int j = 0;
									char* buf = malloc(1);
									buf[0] = 0;
									int counter = 0;
									while(1){
										char ch = value[j];
										if(ch == 0 || ch == '|'){
											if(counter != 0){
												result->p_category = realloc(result->p_category, sizeof(*result->p_category) * (category_count + 2));
												result->p_category[category_count] = sstrdup(buf);
												result->p_category[category_count + 1] = NULL;
												category_count++;
											}
											free(buf);
											buf = malloc(1);
											buf[0] = 0;
											counter = 0;
											if(ch == 0) break;
										}else{
											buf = realloc(buf, counter + 2);
											buf[counter] = ch;
											buf[counter + 1] = 0;
											counter++;
										}
										j++;
									}
									free(buf);
								}else if(strcmp(key, "Bold") == 0){
									CHECK_NULL;
									char* tmp = result->p_body;
									result->p_body = sstrcat4(tmp, "<span class=\"font-bold\">", value, "</span>");
									free(tmp);
								}else if(strcmp(key, "Italic") == 0){
									CHECK_NULL;
									char* tmp = result->p_body;
									result->p_body = sstrcat4(tmp, "<span class=\"font-italic\">", value, "</span>");
									free(tmp);
								}else if(strcmp(key, "BoldItalic") == 0 || strcmp(key, "ItalicBold") == 0){
									CHECK_NULL;
									char* tmp = result->p_body;
									result->p_body = sstrcat4(tmp, "<span class=\"font-bold-italic\">", value, "</span>");
									free(tmp);
								}else if(strcmp(key, "Monospace") == 0){
									CHECK_NULL;
									char* tmp = result->p_body;
									result->p_body = sstrcat4(tmp, "<span class=\"font-monospace\">", value, "</span>");
									free(tmp);
								}else if(strcmp(key, "Center") == 0){
									CHECK_NULL;
									char* tmp = result->p_body;
									result->p_body = sstrcat4(tmp, "<div class=\"style-center\">", value, "</div>");
									free(tmp);
								}else if(strcmp(key, "Left") == 0){
									CHECK_NULL;
									char* tmp = result->p_body;
									result->p_body = sstrcat4(tmp, "<div class=\"style-left\">", value, "</div>");
									free(tmp);
								}else if(strcmp(key, "Right") == 0){
									CHECK_NULL;
									char* tmp = result->p_body;
									result->p_body = sstrcat4(tmp, "<div class=\"style-right\">", value, "</div>");
									free(tmp);
								}else if(strcmp(key, "FloatLeft") == 0){
									CHECK_NULL;
									char* tmp = result->p_body;
									result->p_body = sstrcat4(tmp, "<div class=\"style-fleft\">", value, "</div>");
									free(tmp);
								}else if(strcmp(key, "FloatRight") == 0){
									CHECK_NULL;
									char* tmp = result->p_body;
									result->p_body = sstrcat4(tmp, "<div class=\"style-fright\">", value, "</div>");
									free(tmp);
								}else if(strcmp(key, "Code") == 0){
									code = 1;
									char* tmp = result->p_body;
									result->p_body = sstrcat(tmp, "<pre><code>");
									free(tmp);
								}else if(strcmp(key, "NikkelVersion") == 0){
									char* tmp = result->p_body;
									result->p_body = sstrcat(tmp, NIKKEL_VERSION);
									free(tmp);
								}else if(strcmp(key, "NikkelRevDate") == 0){
									char* tmp = result->p_body;
									char* date = sstrdup(NIKKEL_REVDATE + sstrlen("$Date: "));
									date[sstrlen(date) - 1] = 0;
									result->p_body = sstrcat(tmp, date);
									free(date);
									free(tmp);
								}else if(strcmp(key, "ServerSystem") == 0){
									struct utsname sysinfo;
									uname(&sysinfo);
									char* tmp;
									char* sinfo = sstrcat4(sysinfo.sysname, "/", sysinfo.machine, " ");
									tmp = sinfo;
									sinfo = sstrcat(sinfo, sysinfo.release);
									free(tmp);

									tmp = result->p_body;
									result->p_body = sstrcat(tmp, sinfo);
									free(sinfo);
									free(tmp);
								}else if(strcmp(key, "ServerOS") == 0){
									struct utsname sysinfo;
									uname(&sysinfo);

									char* tmp = result->p_body;
									result->p_body = sstrcat(tmp, sysinfo.sysname);
									free(tmp);
								}else if(strcmp(key, "ServerMachine") == 0){
									struct utsname sysinfo;
									uname(&sysinfo);

									char* tmp = result->p_body;
									result->p_body = sstrcat(tmp, sysinfo.machine);
									free(tmp);
								}else if(strcmp(key, "ServerSoftware") == 0){
									char* tmp = result->p_body;
									result->p_body = sstrcat(tmp, getenv("SERVER_SOFTWARE"));
									free(tmp);
								}else if(strcmp(key, "ServerRelease") == 0){
									struct utsname sysinfo;
									uname(&sysinfo);

									char* tmp = result->p_body;
									result->p_body = sstrcat(tmp, sysinfo.release);
									free(tmp);
								}else if(strcmp(key, "Hostname") == 0){
									char* hostname = malloc(513);
									memset(hostname, 0, 513);
									gethostname(hostname, 512);
									char* tmp = result->p_body;
									result->p_body = sstrcat(tmp, hostname);
									free(tmp);
									free(hostname);
								}else if(strcmp(key, "Align") == 0 || strcmp(key, "/Align") == 0){
									if(strcmp(key, "Align") == 0){
										CHECK_NULL;
										char* align = NULL;
										if(value == NULL){
											align = sstrdup("left");
										}else if(strcmp(value, "Right") == 0 || strcmp(value, "R") == 0){
											align = sstrdup("right");
										}else if(strcmp(value, "Left") == 0 || strcmp(value, "L") == 0){
											align = sstrdup("left");
										}
										char* tmp = align;
										align = sstrcat("special-align-", tmp);
										free(tmp);
										tmp = result->p_body;
										result->p_body = sstrcat4(tmp, "<span class=\"", align, "\">");
										free(tmp);
										free(align);
									}else{
										char* tmp = result->p_body;
										result->p_body = sstrcat(tmp, "</span>");
										free(tmp);
									}
								}else if(strcmp(key, "Float") == 0 || strcmp(key, "/Float") == 0){
									if(strcmp(key, "Float") == 0){
										CHECK_NULL;
										char* align = NULL;
										if(value == NULL){
											align = sstrdup("left");
										}else if(strcmp(value, "Right") == 0 || strcmp(value, "R") == 0){
											align = sstrdup("right");
										}else if(strcmp(value, "Left") == 0 || strcmp(value, "L") == 0){
											align = sstrdup("left");
										}
										char* tmp = align;
										align = sstrcat("special-float-", tmp);
										free(tmp);
										tmp = result->p_body;
										result->p_body = sstrcat4(tmp, "<span class=\"", align, "\">");
										free(tmp);
										free(align);
									}else{
										char* tmp = result->p_body;
										result->p_body = sstrcat(tmp, "</span>");
										free(tmp);
									}
								}else if(strcmp(key, "TablePlace") == 0 || strcmp(key, "TP") == 0){
									CHECK_NULL;
									if(strcmp(value, "Right") == 0 || strcmp(value, "R") == 0){
										table_place = TABLE_RIGHT;	
									}else if(strcmp(value, "Left") == 0 || strcmp(value, "L") == 0){
										table_place = TABLE_LEFT;
									}
								}else if(strcmp(key, "TableWidth") == 0 || strcmp(key, "TW") == 0){
									if(value == NULL){
										if(table_width != NULL) free(table_width);
										table_width = NULL;
									}else{
										char unusable = 0;
										int twi = 0;
										for(twi = 0; value[twi]; twi++){
											char ch = value[twi];
											if(ch == '"'){
												unusable = 1;
												break;
											}
										}
										if(unusable){
											char* err = malloc(513);
											memset(err, 0, 513);
											sprintf(err, "<span class=\"special-message\"><span class=\"color-red\">&emsp;</span><span class=\"special-error\">ERROR</span>: Line number %d: `%s` does not allow using double quotation in the value<span class=\"color-red\">&emsp;</span></span>", line, key);
											char* tmp_body = result->p_body;
											result->p_body = sstrcat(tmp_body, err);
											free(tmp_body);
											free(err);
										}else{
											if(table_width != NULL) free(table_width);
											table_width = sstrdup(value);
										}
									}
								}else if(strcmp(key, "Table") == 0 || strcmp(key, "/Table") == 0){
									if(strcmp(key, "Table") == 0){
										char* place = NULL;
										if(table_place == TABLE_LEFT){
											place = sstrdup("wiki-table-left");
										}else if(table_place == TABLE_RIGHT){
											place = sstrdup("wiki-table-right");
										}
										char* tmp;
										tmp = result->p_body;
										result->p_body = sstrcat4(tmp, "<span class=\"wiki-table-container ", place, "\"");
										free(tmp);
										if(table_width != NULL){
											tmp = result->p_body;
											result->p_body = sstrcat4(tmp, " style=\"width: ", table_width, ";\"");
											free(tmp);
										}
										tmp = result->p_body;
										result->p_body = sstrcat(tmp, ">");
										free(tmp);

										if(value != NULL){
											tmp = result->p_body;
											result->p_body = sstrcat4(tmp, "<div class=\"wiki-table-title\">", value, "</div>");
											free(tmp);
										}
										tmp = result->p_body;
										result->p_body = sstrcat4(tmp, "<table class=\"wiki-table ", place, "\">");
										free(tmp);
									}else if(strcmp(key, "/Table") == 0){
										char* tmp = result->p_body;
										result->p_body = sstrcat(tmp, "</table></span>");
										free(tmp);
									}
								}else if(	strcmp(key, "TableHeader") == 0 || strcmp(key, "/TableHeader") == 0
									||	strcmp(key, "TH") == 0 || strcmp(key, "/TH") == 0){
									if(strcmp(key, "TableHeader") == 0 || strcmp(key, "TH") == 0){
										char* tmp = result->p_body;
										result->p_body = sstrcat(tmp, "<tr class=\"wiki-table-header\">");
										free(tmp);
										types = realloc(types, type_count + 2);
										types[type_count] = TABLE_HEADER;
										types[type_count + 1] = 0;
										type_count++;
									}else if(strcmp(key, "/TableHeader") == 0 || strcmp(key, "/TH") == 0){
										char* tmp = result->p_body;
										result->p_body = sstrcat(tmp, "</tr>");
										free(tmp);
										types = realloc(types, type_count + 1);
										types[type_count] = TABLE_NONE;
										type_count--;
									}
								}else if(	strcmp(key, "TableData") == 0 || strcmp(key, "/TableData") == 0
									||	strcmp(key, "TD") == 0 || strcmp(key, "/TD") == 0){
									if(strcmp(key, "TableData") == 0 || strcmp(key, "TD") == 0){
										char* tmp = result->p_body;
										result->p_body = sstrcat(tmp, "<tr class=\"wiki-table-data\">");
										free(tmp);
										types = realloc(types, type_count + 2);
										types[type_count] = TABLE_DATA;
										types[type_count + 1] = 0;
										type_count++;
									}else if(strcmp(key, "/TableData") == 0 || strcmp(key, "/TD") == 0){
										char* tmp = result->p_body;
										result->p_body = sstrcat(tmp, "</tr>");
										free(tmp);
										types = realloc(types, type_count + 1);
										types[type_count] = TABLE_NONE;
										type_count--;
									}
								}else if(	strcmp(key, "TableRow") == 0 || strcmp(key, "/TableRow") == 0
									||	strcmp(key, "TR") == 0 || strcmp(key, "/TR") == 0){
									if(strcmp(key, "TableRow") == 0 || strcmp(key, "TR") == 0){
										char* tmp = result->p_body;
										result->p_body = sstrcat(tmp, "<tr class=\"wiki-table-row\">");
										free(tmp);
									}else if(strcmp(key, "/TableRow") == 0 || strcmp(key, "/TR") == 0){
										char* tmp = result->p_body;
										result->p_body = sstrcat(tmp, "</tr>");
										free(tmp);
									}
								}else if(	strcmp(key, "TableEntryHeader") == 0 || strcmp(key, "/TableEntryHeader") == 0
									||	strcmp(key, "TEH") == 0 || strcmp(key, "/TEH") == 0){
									if(strcmp(key, "TableEntryHeader") == 0 || strcmp(key, "TEH") == 0){
										char* tmp = result->p_body;
										result->p_body = sstrcat(tmp, "<th class=\"wiki-table-entry\">");
										free(tmp);
									}else if(strcmp(key, "/TableEntryHeader") == 0 || strcmp(key, "/TEH") == 0){
										char* tmp = result->p_body;
										result->p_body = sstrcat(tmp, "</th>");
										free(tmp);
									}
								}else if(	strcmp(key, "TableEntryData") == 0 || strcmp(key, "/TableEntryData") == 0
									||	strcmp(key, "TED") == 0 || strcmp(key, "/TED") == 0){
									if(strcmp(key, "TableEntryData") == 0 || strcmp(key, "TED") == 0){
										char* tmp = result->p_body;
										result->p_body = sstrcat(tmp, "<td class=\"wiki-table-entry\">");
										free(tmp);
									}else if(strcmp(key, "/TableEntryData") == 0 || strcmp(key, "/TED") == 0){
										char* tmp = result->p_body;
										result->p_body = sstrcat(tmp, "</td>");
										free(tmp);
									}
								}else if(	strcmp(key, "TableEntry") == 0 || strcmp(key, "/TableEntry") == 0
									||	strcmp(key, "TE") == 0 || strcmp(key, "/TE") == 0){
									char* tag = NULL;
									if(types[type_count - 1] == TABLE_HEADER){
										tag = sstrdup("th");
									}else if(types[type_count - 1] == TABLE_DATA){
										tag = sstrdup("td");
									}
									if(strcmp(key, "TableEntry") == 0 || strcmp(key, "TE") == 0){
										char* tmp = result->p_body;
										result->p_body = sstrcat4(tmp, "<", tag, " class=\"wiki-table-entry\">");
										free(tmp);
									}else if(strcmp(key, "/TableEntry") == 0 || strcmp(key, "/TE") == 0){
										char* tmp = result->p_body;
										result->p_body = sstrcat4(tmp, "</", tag, ">");
										free(tmp);
									}
									free(tag);
								}else if(strcmp(key, "SectionListPlace") == 0){
									CHECK_NULL;
									if(strcmp(value, "Right") == 0 || strcmp(value, "R") == 0){
										result->p_section = SECTION_RIGHT;	
									}else if(strcmp(value, "Left") == 0 || strcmp(value, "L") == 0){
										result->p_section = SECTION_LEFT;
									}
								}else if(strcmp(key, "Forbid") == 0){
									forbid = 1;
								}else if(strcmp(key, "Comment") == 0){
								}
								if(value != NULL) free(value);
							}
clean_sbracket:
							sbracket = 0;
							free(sbracket_buffer);
							sbracket_buffer = malloc(1);
							sbracket_buffer[0] = 0;
							if(forbid == 1){
								result->p_forbid = 1;
								break;
							}
						}else{
							if(ch[0] == '['){
								bracket_count++;
							}else if(ch[0] == ']'){
								bracket_count--;
							}
							char* tmp = sbracket_buffer;
							sbracket_buffer = sstrcat(tmp, ch);
							free(tmp);
						}
						continue;
					}
					if(ch[0] == '['){
						sbracket = 1;
						start_sbracket = i;
					}else{
						char* tmp = result->p_body;
						if(ch[0] == '<'){
							result->p_body = sstrcat(tmp, "&lt;");
						}else if(ch[0] == '>'){
							result->p_body = sstrcat(tmp, "&gt;");
						}else if(ch[0] == '{' && config->c_allow_rawhtml == 1){
							result->p_body = sstrcat(tmp, "<");
						}else if(ch[0] == '}' && config->c_allow_rawhtml == 1){
							result->p_body = sstrcat(tmp, ">");
						}else if(ch[0] == '&'){
							result->p_body = sstrcat(tmp, "&amp;");
						}else if(ch[0] == '\\'){
							if(buffer[i + 1] != 0){
								ch[0] = buffer[i + 1];
								result->p_body = sstrcat(tmp, ch);
								i++;
							}else{
								char* err = malloc(513);
								memset(err, 0, 513);
								sprintf(err, "<span class=\"special-message\"><span class=\"color-yellow\">&emsp;</span><span class=\"special-warn\">WARN</span>: Line number %d: Stray backslash found at the end of the line.<span class=\"color-yellow\">&emsp;</span></span><br>", line);
								result->p_body = sstrcat(tmp, err);
								free(err);
							}
						}else{
							result->p_body = sstrcat(tmp, ch);
						}
						free(tmp);
					}
				}
				free(sbracket_buffer);
				free(ch);
			}else{
				char* tmp = result->p_body;
				if(code == 1){
					result->p_body = sstrcat(tmp, "\n");
				}else{
					result->p_body = sstrcat(tmp, "<br>\n");
				}
				free(tmp);
			}
			if(counter != 0 && code == 1){
				char* tmp = result->p_body;
				result->p_body = sstrcat(tmp, "\n");
				free(tmp);
			}
			free(buffer);
			buffer = malloc(1);
			buffer[0] = 0;
			counter = 0;
			if(f == NULL ? ch == 0 : feof(f)) break;
		}else{
			buffer = realloc(buffer, counter + 2);
			buffer[counter] = ch;
			buffer[counter + 1] = 0;
			counter++;
		}
	}
	free(buffer);
	if(sections[0] != NULL){
		char* str= malloc(1);
		str[0] = 0;
		char* tmp;
		tmp = str;
		char* classname = NULL;
		if(result->p_section == SECTION_LEFT){
			classname = sstrdup("section-left");
		}else if(result->p_section == SECTION_RIGHT){
			classname = sstrdup("section-right");
		}
		str = sstrcat4(tmp, "<span class=\"section-container ", classname, "\">");
		free(classname);
		free(tmp);
		tmp = str;
		str = sstrcat(tmp, "<div id=\"section-title\">Sections</div>");
		free(tmp);

		tmp = str;
		str = sstrcat(tmp, "<pre>");
		free(tmp);

		tmp = str;
		str = sstrcat(tmp, "<code>");
		free(tmp);

		for(i = 0; sections[i] != 0; i++){
			char* num = malloc(513);
			memset(num, 0, 513);
			sprintf(num, "%d.   ", i + 1);

			char* fmt = sstrcat(num, sections[i]->name);
			free(num);
			
			num = malloc(513);
			memset(num, 0, 513);
			sprintf(num, "%d-", i + 1);

			char* replaced = sreplace(sections[i]->name, ' ', '-');

			char* start = sstrcat5("<a href=\"#section-", num, "0-", replaced, "\">");

			tmp = str;
			str = sstrcat5(tmp, start, fmt, "</a>", "\n");
			free(tmp);
			free(num);
			free(replaced);

			free(start);

			free(fmt);

			int j;
			for(j = 0; sections[i]->subsections[j] != NULL; j++){
				num = malloc(513);
				memset(num, 0, 513);
				sprintf(num, "%d.%d. ", i + 1, j + 1);

				fmt = sstrcat(num, sections[i]->subsections[j]);
				free(num);

				num = malloc(513);
				memset(num, 0, 513);
				sprintf(num, "%d-%d", i + 1, j + 1);

				replaced = sreplace(sections[i]->subsections[j], ' ', '-');

				start = sstrcat5("<a href=\"#section-", num, "-", replaced, "\">");

				tmp = str;
				str = sstrcat5(tmp, start, fmt, "</a>", "\n");
				free(tmp);
				free(num);
				free(replaced);

				free(start);

				free(fmt);

				free(sections[i]->subsections[j]);
			}
			free(sections[i]->subsections);
			free(sections[i]->name);
			free(sections[i]);
		}

		tmp = str;
		str = sstrcat(tmp, "</code>");
		free(tmp);

		tmp = str;
		str = sstrcat(tmp, "</pre>");
		free(tmp);

		tmp = str;
		str = sstrcat(tmp, "</span>");
		free(tmp);
		tmp = result->p_body;
		result->p_body = sstrcat(str, tmp);
		free(tmp);
		free(str);

	}
	if(table_width != NULL) free(table_width);
	free(sections);
	free(types);
	return result;
}

void nikkel_parsed_free(struct nikkel_parsed* parsed){
	free(parsed->p_title);
	free(parsed->p_body);
	int i;
	for(i = 0; parsed->p_category[i] != NULL; i++){
		free(parsed->p_category[i]);
	}
	free(parsed->p_category);
	if(parsed->p_pagetitle != NULL) free(parsed->p_pagetitle);
	free(parsed);
}

void nikkel_dir(const char* dirpath, struct nikkel_entry*** entries, int* entry_count, const char* category, const char* prefix, const char* entryprefix){
	DIR* dir = opendir(dirpath);
	if(dir != NULL){
		struct dirent* d;
		while((d = readdir(dir)) != NULL){
			if(d->d_name[0] == '.'){
				continue;
			}
			char* stat_path = sstrcat3(dirpath, "/", d->d_name);
			struct stat s;
			stat(stat_path, &s);
			free(stat_path);
			if(S_ISDIR(s.st_mode)){
				stat_path = sstrcat3(dirpath, "/", d->d_name);
				char* dir_prefix = sstrcat3(prefix, "/", d->d_name);
				char* dir_entryprefix = sstrcat3(entryprefix, d->d_name, "/");
				nikkel_dir(stat_path, entries, entry_count, category, dir_prefix, dir_entryprefix);
				free(dir_entryprefix);
				free(dir_prefix);
				free(stat_path);
			}else if(S_ISREG(s.st_mode)){
				char* fmt_name = sstrdup(d->d_name);
				fmt_name[sstrlen(fmt_name) - sstrlen(".wiki")] = 0;
				char* name = sstrdup(fmt_name);
				free(fmt_name);
				char* orig_name = sstrdup(name);
				char** categories = malloc(sizeof(*categories));
				categories[0] = NULL;
				int category_count = 0;
				char* path = sstrcat3(dirpath, "/", d->d_name);
				FILE* f = fopen(path, "r");
				if(f != NULL){
					int counter = 0;
					char* buffer = malloc(1);
					buffer[0] = 0;
					char ch;
					while(1){
						fread(&ch, 1, 1, f);
						if(feof(f) || ch == '\n'){
							if(counter != 0){
								int i;
								char sbracket = 0;
								char* chb = malloc(2);
								chb[0] = 0;
								chb[1] = 0;
								char* sbracket_buffer = malloc(1);
								sbracket_buffer[0] = 0;
								for(i = 0; buffer[i] != 0; i++){
									chb[0] = buffer[i];
									if(sbracket == 1){
										if(chb[0] == ']'){
											size_t ind = sindexof(sbracket_buffer, "|");
											sbracket_buffer[ind == -1 ? sstrlen(sbracket_buffer) : ind] = 0;

											char* key = sbracket_buffer;
											char* value = sbracket_buffer + (ind == -1 ? 0 : (ind + 1));
											if(strcmp(key, "Category") == 0){
												int j = 0;
												char* category_buffer = malloc(1);
												category_buffer[0] = 0;
												int category_bcounter = 0;
												while(1){
													if(value[j] == 0 || value[j] == '|'){
														if(category_bcounter != 0){
															categories = realloc(categories, sizeof(*categories) * (category_count + 2));
															categories[category_count] = sstrdup(category_buffer);
															categories[category_count + 1] = NULL;
															category_count++;
														}
														free(category_buffer);
														category_buffer = malloc(1);
														category_buffer[0] = 0;
														category_bcounter = 0;
														if(value[j] == 0) break;
													}else{
														category_buffer = realloc(category_buffer, category_bcounter + 2);
														category_buffer[category_bcounter] = value[j];
														category_buffer[category_bcounter + 1] = 0;
														category_bcounter++;
													}
													j++;
												}
												free(category_buffer);
											}else if(strcmp(key, "Title") == 0){
												free(name);
												name = sstrdup(value);
											}
											sbracket = 0;
											free(sbracket_buffer);
											sbracket_buffer = malloc(1);
											sbracket_buffer[0] = 0;
										}else{
											char* tmp = sbracket_buffer;
											sbracket_buffer = sstrcat(tmp, chb);
											free(tmp);
										}
										continue;
									}
									if(chb[0] == '['){
										sbracket = 1;
									}
								}
							}
							free(buffer);
							buffer = malloc(1);
							buffer[0] = 0;
							counter = 0;
							if(feof(f)) break;
						}else{
							buffer = realloc(buffer, counter + 2);
							buffer[counter] = ch;
							buffer[counter + 1] = 0;
							counter++;
						}
					}
					free(buffer);
					fclose(f);
				}
				free(path);
				int i;
				char matched = 0;
				for(i = 0; categories[i] != NULL; i++){
					if(strcmp(categories[i], category) == 0){
						matched = 1;
						break;
					}
				}
				if(matched == 1){
					char letter = '*';
					if('a' <= name[0] && name[0] <= 'z'){
						letter = name[0] - 0x20;
					}else if('A' <= name[0] && name[0] <= 'Z'){
						letter = name[0];
					}
					*entries = realloc(*entries, sizeof(**entries) * ((*entry_count) + 2));
					struct nikkel_entry* ent = malloc(sizeof(*ent));
					ent->name = sstrdup(name);
					ent->original_name = sstrdup(orig_name);
					ent->letter = letter;
					ent->prefix = sstrdup(prefix);
					ent->entryprefix = sstrdup(entryprefix);
					(*entries)[(*entry_count)] = ent;
					(*entries)[(*entry_count) + 1] = NULL;
					(*entry_count)++;
				}
				for(i = 0; categories[i] != NULL; i++) free(categories[i]);
				free(categories);
				free(name);
				free(orig_name);
			}
		}
		closedir(dir);
	}

}

char* nikkel_parse_category(struct nikkel_config* config, char* category){
	char* result = malloc(1);
	result[0] = 0;
	struct nikkel_entry** entries = malloc(sizeof(*entries));
	entries[0] = 0;
	int entry_count = 0;
	nikkel_dir(config->c_wikipath, &entries, &entry_count, category, config->c_prefix, "");
	int i;
	char ch = 'A';
	char match = 0;
	int matched_count = 0;
	for(ch = 'A'; ch != '['; ch++){
		int j;
		char first = 0;
		char* tmp;
		for(i = 0; entries[i] != NULL; i++){
			if(entries[i]->letter == ch){
				matched_count++;
				match = 1;
				if(first == 0){
					char* chb = malloc(2);
					chb[0] = ch;
					chb[1] = 0;
					tmp = result;
					result = sstrcat4(tmp, "<h3 id=\"char-", chb, "\"><span>");
					free(tmp);
					tmp = result;
					result = sstrcat(tmp, chb);
					free(tmp);
					tmp = result;
					result = sstrcat4(tmp, "</span> <a href=\"#char-", chb, "\">&#167;</a></h3>");
					free(tmp);
					free(chb);

					tmp = result;
					result = sstrcat(tmp, "<ul>");
					free(tmp);

					first = 1;
				}
				char* tmp;
				tmp = result;
				result = sstrcat4(tmp, "<li><a href=\"", entries[i]->prefix, "/");
				free(tmp);

				tmp = result;
				result = sstrcat4(tmp, entries[i]->original_name, "\">", entries[i]->entryprefix);
				free(tmp);

				tmp = result;
				result = sstrcat(tmp, entries[i]->name);
				free(tmp);

				tmp = result;
				result = sstrcat(tmp, "</a></li>");
				free(tmp);
			}
		}
		if(first == 1){
			tmp = result;
			result = sstrcat(tmp, "</ul>");
			free(tmp);
		}
	}
	for(i = 0; entries[i] != NULL; i++){	
		free(entries[i]->name);
		free(entries[i]->original_name);
		free(entries[i]->prefix);
		free(entries[i]->entryprefix);
		free(entries[i]);
	}
	free(entries);
	if(match == 0){
		free(result);
		result = sstrdup("Couldn't find the article which is in the category.");
	}else{
		char* matched_str = malloc(513);
		memset(matched_str, 0, 513);
		sprintf(matched_str, "There are %d articles which are in the category.<hr>", matched_count);
		char* tmp = result;
		result = sstrcat(matched_str, tmp);
		free(matched_str);
		free(tmp);
	}
	return result;
}
